#include <iostream>
#include <vector>
#include <functional>
#include <SDL2/SDL.h>
#include <SDL2/SDL_ttf.h>
#include "Config.h"
using namespace std;

#include "World.h"
#include "Random.h"
#include "Menu.h"

// Stores all global variables. Uses Singleton Design Pattern to make sure only one object of this class is constructed
class Globals {
    public:

    // Deletes copy constructor and assignment operator to prevent construction of more than one object
    Globals(const Globals&) = delete; // delete Copy constructor
    Globals& operator=(const Globals&) = delete; // delete Copy assignment operator

    // Static method to access the singleton object/instance.
    static Globals& getInstance() {
        static Globals instance; // Guaranteed to be destroyed and instantiated on first use
        return instance;
    }

    // Set sizes of things
    const int grid_cell_width = 250;
    const int grid_cell_height = 250;
    const int window_cell_width = 80;
    const int window_cell_height = 80;
    const int panel_width = 240;
    const int mini_map_height = 240;
    const int mini_map_cell_margin = 10;

    // Set frame rate
    const int FPS = 40;
    const int frameDelay = 1000 / FPS;

    // SDL objects
    SDL_Window* window = nullptr;
    SDL_Renderer* renderer = nullptr;
    TTF_Font* font = nullptr;
    bool running = true;

    // Minimap coordinate
    int minimap_x = 0;
    int minimap_y = 0;

    // Mouse state
    int mouseX = 0;
    int mouseY = 0;
    bool mouse_button_left_down = false;
    bool mouse_button_right_down = false;
    bool mouse_button_left_click = false;
    bool mouse_button_right_click = false;

    private:

    // Private constructor to prevent instancing from outside the class.
    Globals() {}
};

// Create texture from text
SDL_Texture* create_texture(string text, SDL_Color& color, TTF_Font* font, SDL_Renderer* renderer)
{
    SDL_Surface* surface = TTF_RenderText_Solid(font, text.c_str(), color);
    SDL_Texture* texture = SDL_CreateTextureFromSurface(renderer, surface);
    SDL_FreeSurface(surface);
    return texture;
}

// Stores a texture and callback for a Button
class Button {
    public:
        Button(SDL_Texture* texture, std::function<void()> callback) :texture(texture),callback(callback) 
        {}

        SDL_Texture* get_texture() {
            return texture;
        }

        void operator()() {
            callback();
        }

    private:
        SDL_Texture* texture;
        std::function<void()> callback;  
};

// Randomizing the grid
void randomize_grid(vector< vector<bool> >& grid)
{
    for (int y=0;y<int(grid.size());y++) {
        for (int x=0;x<int(grid[y].size());x++) {
            grid[y][x] = rand() % 2;
        }
    }
}

// Clear the grid
void clear_grid(vector< vector<bool> >& grid)
{
    for (int y=0;y<int(grid.size());y++) {
        for (int x=0;x<int(grid[y].size());x++) {
            grid[y][x] = false;
        }
    }
}

// Load a pattern into the grid
void load_grid(vector< vector<bool> >& grid)
{
    string filename;
    cout<<"Type file to load: ";
    cin>> filename;
    cout<<"Loading file '"<<filename<<"'"<<endl;
    // for now do nothing and just clear the grid
    clear_grid(grid);
}

// Initializes the SDL window, renderer, and font
bool init_SDL(Globals& globals)
{
    // Init SDL 
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        SDL_Log("SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
    }

    if (TTF_Init() == -1) {
        SDL_Log("TrueType Font (TTF) could not be created! SDL_Error: %s\n", SDL_GetError());
        SDL_Quit();
        return false;
    }

    // Create window
    globals.window = SDL_CreateWindow("Game of Life",
                                            SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                            800, 600,
                                            SDL_WINDOW_SHOWN | SDL_WINDOW_RESIZABLE);
    if (!globals.window) {
        SDL_Log("Window could not be created! SDL_Error: %s\n", SDL_GetError());
        SDL_Quit();
        return false;
    }

    // Create renderer
    globals.renderer = SDL_CreateRenderer(globals.window, -1, SDL_RENDERER_ACCELERATED);
    if (!globals.renderer) {
        SDL_Log("Renderer could not be created! SDL Error: %s\n", SDL_GetError());
        SDL_DestroyWindow(globals.window);
        SDL_Quit();
        return false;
    }

    // Load TrueType Font
    string font_name = string(CMAKE_SOURCE_DIR) + "/Comic_Sans_MS.ttf";
    globals.font = TTF_OpenFont(font_name.c_str(), 24);
    if (globals.font == nullptr) {
        SDL_DestroyRenderer(globals.renderer);
        SDL_DestroyWindow(globals.window);
        SDL_Log("TTF_OpenFont Error: %s\n", TTF_GetError());
        TTF_Quit();
        SDL_Quit();
        return false;
    }
    return true;
}

// Create each buttons and set its callback to manipulate the grid
vector<Button> create_buttons(Globals& globals, vector< vector<bool> >& grid)
{
    // Create buttons
    SDL_Color color = {255, 255, 255, 255}; // White
    vector<Button> buttons;
    buttons.push_back(Button(create_texture("Randomize", color, globals.font, globals.renderer), 
                            [&](){
                                cout<<"Randomize"<<endl; 
                                randomize_grid(grid); 
                            }
                            ));
    buttons.push_back(Button(create_texture("Clear", color, globals.font, globals.renderer), 
                            [&](){
                                cout<<"Clear"<<endl; 
                                clear_grid(grid); 
                            }
                            ));
    buttons.push_back(Button(create_texture("Load", color, globals.font, globals.renderer), 
                            [&](){
                                cout<<"Load"<<endl;
                                load_grid(grid);
                            }
                            ));
    return buttons;
}

// Handles the SDL events setting the running state, mouse position and mouse button states
void handle_events(Globals& globals)
{
    SDL_Event event;
    globals.mouse_button_left_click = false; // set clicks to false
    globals.mouse_button_right_click = false;
    while (SDL_PollEvent(&event)) {
        switch (event.type) {
        case SDL_QUIT:
            globals.running = false;
            break;
        case SDL_MOUSEMOTION: // deals with moue movement
            globals.mouseX = event.motion.x;
            globals.mouseY = event.motion.y;
            break;
        case SDL_MOUSEBUTTONDOWN: // deals with moue button presses
            switch (event.button.button) {
                case SDL_BUTTON_LEFT:
                    globals.mouse_button_left_down = true;
                    globals.mouse_button_left_click = true;
                    break;
                case SDL_BUTTON_RIGHT:
                    globals.mouse_button_right_down = true;
                    globals.mouse_button_right_click = true;
                    break;
                default:
                    break;
            }
            break;
        case SDL_MOUSEBUTTONUP: // deals with moue button releases
            switch (event.button.button) {
                case SDL_BUTTON_LEFT:
                    globals.mouse_button_left_down = false;
                    break;
                case SDL_BUTTON_RIGHT:
                    globals.mouse_button_right_down = false;
                    break;
                default:
                    break;
            }
            break;
        }
    }
}

// Draws the minimap and changes the globals.minimap_x and globals.minimap_y variables that determine what part of the world is shown
void draw_minimap_and_handle_clicks(Globals& globals)
{
    SDL_Rect mini_map_rect = {0, 0, globals.panel_width, globals.mini_map_height};
    double minimap_scale_width = double(globals.panel_width) /  (globals.grid_cell_width + globals.mini_map_cell_margin * 2);
    double minimap_scale_height = double(globals.mini_map_height) / (globals.grid_cell_height + globals.mini_map_cell_margin * 2);
    // When left mouse button is pressed set the minimap position based on mouse position
    if (globals.mouse_button_left_down && globals.mouseX < globals.panel_width && globals.mouseY < globals.mini_map_height) {
        globals.minimap_x = globals.mouseX / minimap_scale_width - globals.window_cell_width/2;
        globals.minimap_y = globals.mouseY / minimap_scale_height - globals.window_cell_height/2;
        globals.minimap_x = max(0, min(globals.grid_cell_width + globals.mini_map_cell_margin*2 - globals.window_cell_width, globals.minimap_x));
        globals.minimap_y = max(0, min(globals.grid_cell_height + globals.mini_map_cell_margin*2 - globals.window_cell_height, globals.minimap_y));
    }
    // Draw the minimap
    SDL_RenderDrawRect(globals.renderer, &mini_map_rect);
    SDL_Rect grid_rect = {int(globals.mini_map_cell_margin * minimap_scale_width),
                        int(globals.mini_map_cell_margin * minimap_scale_height),
                        int(globals.grid_cell_width * minimap_scale_width),
                        int(globals.grid_cell_height * minimap_scale_height)};
    SDL_RenderDrawRect(globals.renderer, &grid_rect);
    SDL_Rect window_rect = {int(globals.minimap_x * minimap_scale_width), 
                            int(globals.minimap_y * minimap_scale_height), 
                            int(globals.window_cell_width * minimap_scale_width), 
                            int(globals.window_cell_height * minimap_scale_height)};
    SDL_RenderDrawRect(globals.renderer, &window_rect);
}

// Draws each button and calls the callback of the button that is clicked by the mouse 
void draw_buttons_and_handle_clicks(Globals &globals, vector<Button>& buttons)
{
    for (int i=0; i<int(buttons.size()); i++) {
        SDL_Rect dstrect = {0, globals.mini_map_height + i*50, globals.panel_width, 50};
        SDL_RenderCopy(globals.renderer, buttons[i].get_texture(), NULL, &dstrect);
        // test if mouse is on the button
        if (globals.mouseX >= dstrect.x && globals.mouseX <= dstrect.x + dstrect.w &&
            globals.mouseY >= dstrect.y && globals.mouseY <= dstrect.y + dstrect.h) {
            SDL_RenderDrawRect(globals.renderer, &dstrect);
            if (globals.mouse_button_left_click) {
                buttons[i](); // call callback function associate with button
            }
        }
    }
}

// Draws the grid and changes the state of a cell when the mouse is on it and mouse button is down 
void draw_grid_and_handle_clicks(Globals &globals,vector< vector<bool> >& grid, int window_width, int window_height)
{
    // scale the cell size based on the window size
    double cell_width  = static_cast<double>(window_width-globals.panel_width) / globals.window_cell_width;
    double cell_height = static_cast<double>(window_height) / globals.window_cell_height;
    // x,y of cell that is in the top left of the window
    int top_x = globals.minimap_x - globals.mini_map_cell_margin;
    int top_y = globals.minimap_y - globals.mini_map_cell_margin;
    // mouse coordinate as cell indices in the windows
    int mouse_cell_x = (globals.mouseX-globals.panel_width)/cell_width;
    int mouse_cell_y = globals.mouseY/cell_height;
    if (globals.mouseX> globals.panel_width) {
        // mouse coordinate as cell indices in the grid
        int mouse_grid_x = top_x + mouse_cell_x;
        int mouse_grid_y = top_y + mouse_cell_y;
        // change cell when mouse button is down
        if (mouse_grid_x>=0 && mouse_grid_x<globals.grid_cell_width && 
            mouse_grid_y>=0 && mouse_grid_y<globals.grid_cell_height) {
            if (globals.mouse_button_left_down) {
                grid[mouse_grid_y][mouse_grid_x] = true;
            }
            if (globals.mouse_button_right_down) {
                grid[mouse_grid_y][mouse_grid_x] = false;
            }
        }
    }
    // for each cell in window 
    for (int y=0; y<globals.window_cell_height; y++) {
        int cell_y = top_y + y;
        if (cell_y>=0 && cell_y<globals.grid_cell_height) {
            for (int x=0; x<globals.window_cell_width; x++) {
                int cell_x = top_x + x;
                if (cell_x>=0 && cell_x<globals.grid_cell_width) {
                    // draw the cell
                    SDL_Rect rect = {int(x * cell_width) + globals.panel_width, 
                                    int(y * cell_height), 
                                    int(cell_width)-2,
                                    int(cell_height)-2};
                    if (x==mouse_cell_x && y==mouse_cell_y)
                        SDL_SetRenderDrawColor(globals.renderer, 255, 0, 0, 255); // red
                    else
                        SDL_SetRenderDrawColor(globals.renderer, 255, 255, 255, 255); // white
                    if (grid[cell_y][cell_x]) {
                        SDL_RenderFillRect(globals.renderer, &rect);
                    }
                    else {
                        SDL_RenderDrawRect(globals.renderer, &rect);
                    }
                }
            }
        }
    }
}

int main(int argc, char* argv[])
{
    Globals& globals = Globals::getInstance();

    if (!init_SDL(globals))
        return 1;

    // Create grid
    vector< vector<bool> > grid(globals.grid_cell_height, vector<bool>(globals.grid_cell_width, false));
    randomize_grid(grid);

    vector<Button> buttons = create_buttons(globals,grid);
    
    // Main loop
    Uint32 frameStart;
    int frameTime;
    while (globals.running) {
        frameStart = SDL_GetTicks();
        
        // Handle events
        handle_events(globals);
        
        // Clear the window
        SDL_SetRenderDrawColor(globals.renderer, 0, 0, 0, 255); // black
        SDL_RenderClear(globals.renderer);

        // Get current window size
        int window_width, window_height;
        SDL_GetWindowSize(globals.window, &window_width, &window_height);
        SDL_SetRenderDrawColor(globals.renderer, 255, 255, 255, 255); // white

        // Draw Mini Map
        draw_minimap_and_handle_clicks(globals);

        // Draw Buttons
        draw_buttons_and_handle_clicks(globals, buttons);

        // Draw grid
        draw_grid_and_handle_clicks(globals, grid, window_width, window_height);

        // Update window with rendering and control frame rate 
        SDL_RenderPresent(globals.renderer);
        frameTime = SDL_GetTicks() - frameStart;
        if (globals.frameDelay > frameTime) {
            SDL_Delay(globals.frameDelay - frameTime);
        }
    }

    // Clean up
    SDL_DestroyRenderer(globals.renderer);
    SDL_DestroyWindow(globals.window);
    SDL_Quit();
    return 0;
}
